from .terms import PseudoColumn

ColumnValue = PseudoColumn("COLUMN_VALUE")
ObjectID = PseudoColumn("OBJECT_ID")
ObjectValue = PseudoColumn("OBJECT_VALUE")
RowNum = PseudoColumn("ROWNUM")
RowID = PseudoColumn("ROWID")
SysDate = PseudoColumn("SYSDATE")
